@@ -3,13 +3,21 @@
 # Thomas Nagy, 2006-2010 (ita)
 
 """
-Support for the Qt4 libraries and tools::
+
+Tool Description
+================
+
+This tool helps with finding Qt4 tools and libraries,
+and also provides syntactic sugar for using Qt4 tools.
+
+The following snippet illustrates the tool usage::
 
 	def options(opt):
 		opt.load('compiler_cxx qt4')
+
 	def configure(conf):
 		conf.load('compiler_cxx qt4')
-		conf.env.append_value('CXXFLAGS', ['-g']) # test
+
 	def build(bld):
 		bld(
 			features = 'qt4 cxx cxxprogram',
@@ -18,23 +26,41 @@ Support for the Qt4 libraries and tools::
 			target   = 'window',
 		)
 
-The C++ files must include the .moc files, which is regarded as the
-best practice (much faster compilations). This also implies that the
-include paths have to be set properly. To have the include paths added
-automatically, use the following::
+Here, the UI description and resource files will be processed
+to generate code.
 
-	from waflib.TaskGen import feature, before_method, after_method
-	@feature('cxx')
-	@after_method('process_source')
-	@before_method('apply_incpaths')
-	def add_includes_paths(self):
-		incs = set(self.to_list(getattr(self, 'includes', '')))
-		for x in self.compiled_tasks:
-			incs.add(x.inputs[0].parent.path_from(self.path))
-		self.includes = list(incs)
+Usage
+=====
+
+Load the "qt4" tool.
+
+You also need to edit your sources accordingly:
+
+- the normal way of doing things is to have your C++ files
+  include the .moc file.
+  This is regarded as the best practice (and provides much faster
+  compilations).
+  It also implies that the include paths have beenset properly.
+
+- to have the include paths added automatically, use the following::
+
+     from waflib.TaskGen import feature, before_method, after_method
+     @feature('cxx')
+     @after_method('process_source')
+     @before_method('apply_incpaths')
+     def add_includes_paths(self):
+        incs = set(self.to_list(getattr(self, 'includes', '')))
+        for x in self.compiled_tasks:
+            incs.add(x.inputs[0].parent.path_from(self.path))
+        self.includes = list(incs)
+
+Note: another tool provides Qt processing that does not require
+.moc includes, see 'playground/slow_qt/'.
+
+A few options (--qt{dir,bin,...}) and environment variables
+(QT4_{ROOT,DIR,MOC,UIC,XCOMPILE}) allow finer tuning of the tool,
+tool path selection, etc; please read the source for more info.
 
-Another tool provides a Qt processing that does not require the moc
-includes. See http://code.google.com/p/waf/source/browse/trunk/playground/slow_qt/
 """
 
 try:
@@ -47,8 +73,8 @@ else:
 	has_xml = True
 
 import os, sys
-from waflib.Tools import c_preproc, cxx
-from waflib import Task, Utils, Options, Errors
+from waflib.Tools import cxx
+from waflib import Task, Utils, Options, Errors, Context
 from waflib.TaskGen import feature, after_method, extension
 from waflib.Configure import conf
 from waflib import Logs
@@ -73,9 +99,9 @@ EXT_QT4 = ['.cpp', '.cc', '.cxx', '.C']
 File extensions of C++ files that may require a .moc processing
 """
 
-QT4_LIBS = "QtCore QtGui QtUiTools QtNetwork QtOpenGL QtSql QtSvg QtTest QtXml QtXmlPatterns QtWebKit Qt3Support QtHelp QtScript QtDeclarative"
+QT4_LIBS = "QtCore QtGui QtUiTools QtNetwork QtOpenGL QtSql QtSvg QtTest QtXml QtXmlPatterns QtWebKit Qt3Support QtHelp QtScript QtDeclarative QtDesigner"
 
-class qxx(cxx.cxx):
+class qxx(Task.classes['cxx']):
 	"""
 	Each C++ file can have zero or several .moc files to create.
 	They are known only when the files are scanned (preprocessor)
@@ -88,16 +114,6 @@ class qxx(cxx.cxx):
 		Task.Task.__init__(self, *k, **kw)
 		self.moc_done = 0
 
-	def scan(self):
-		"""Re-use the C/C++ scanner, but remove the moc files from the dependencies"""
-		(nodes, names) = c_preproc.scan(self)
-		# for some reasons (variants) the moc node may end in the list of node deps
-		for x in nodes:
-			if x.name.endswith('.moc'):
-				nodes.remove(x)
-				names.append(x.path_from(self.inputs[0].parent.get_bld()))
-		return (nodes, names)
-
 	def runnable_status(self):
 		"""
 		Compute the task signature to make sure the scanner was executed. Create the
@@ -113,6 +129,47 @@ class qxx(cxx.cxx):
 			self.add_moc_tasks()
 			return Task.Task.runnable_status(self)
 
+	def create_moc_task(self, h_node, m_node):
+		"""
+		If several libraries use the same classes, it is possible that moc will run several times (Issue 1318)
+		It is not possible to change the file names, but we can assume that the moc transformation will be identical,
+		and the moc tasks can be shared in a global cache.
+
+		The defines passed to moc will then depend on task generator order. If this is not acceptable, then
+		use the tool slow_qt4 instead (and enjoy the slow builds... :-( )
+		"""
+		try:
+			moc_cache = self.generator.bld.moc_cache
+		except AttributeError:
+			moc_cache = self.generator.bld.moc_cache = {}
+
+		try:
+			return moc_cache[h_node]
+		except KeyError:
+			tsk = moc_cache[h_node] = Task.classes['moc'](env=self.env, generator=self.generator)
+			tsk.set_inputs(h_node)
+			tsk.set_outputs(m_node)
+
+			if self.generator:
+				self.generator.tasks.append(tsk)
+
+			# direct injection in the build phase (safe because called from the main thread)
+			gen = self.generator.bld.producer
+			gen.outstanding.insert(0, tsk)
+			gen.total += 1
+
+			return tsk
+
+	def moc_h_ext(self):
+		ext = []
+		try:
+			ext = Options.options.qt_header_ext.split()
+		except AttributeError:
+			pass
+		if not ext:
+			ext = MOC_H
+		return ext
+
 	def add_moc_tasks(self):
 		"""
 		Create the moc tasks by looking in ``bld.raw_deps[self.uid()]``
@@ -130,82 +187,54 @@ class qxx(cxx.cxx):
 			# remove the signature, it must be recomputed with the moc task
 			delattr(self, 'cache_sig')
 
-		moctasks=[]
-		mocfiles=[]
-		try:
-			tmp_lst = bld.raw_deps[self.uid()]
-			bld.raw_deps[self.uid()] = []
-		except KeyError:
-			tmp_lst = []
-		for d in tmp_lst:
+		include_nodes = [node.parent] + self.generator.includes_nodes
+
+		moctasks = []
+		mocfiles = set([])
+		for d in bld.raw_deps.get(self.uid(), []):
 			if not d.endswith('.moc'):
 				continue
-			# paranoid check
-			if d in mocfiles:
-				Logs.error("paranoia owns")
-				continue
+
 			# process that base.moc only once
-			mocfiles.append(d)
-
-			# find the extension - this search is done only once
+			if d in mocfiles:
+				continue
+			mocfiles.add(d)
 
+			# find the source associated with the moc file
 			h_node = None
-			try: ext = Options.options.qt_header_ext.split()
-			except AttributeError: pass
-			if not ext: ext = MOC_H
 
 			base2 = d[:-4]
-			for x in [node.parent] + self.generator.includes_nodes:
-				for e in ext:
+			for x in include_nodes:
+				for e in self.moc_h_ext():
 					h_node = x.find_node(base2 + e)
 					if h_node:
 						break
-				else:
-					continue
-				break
+				if h_node:
+					m_node = h_node.change_ext('.moc')
+					break
 			else:
-				raise Errors.WafError('no header found for %r which is a moc file' % d)
+				# foo.cpp -> foo.cpp.moc
+				for k in EXT_QT4:
+					if base2.endswith(k):
+						for x in include_nodes:
+							h_node = x.find_node(base2)
+							if h_node:
+								break
+						if h_node:
+							m_node = h_node.change_ext(k + '.moc')
+							break
 
-			# next time we will not search for the extension (look at the 'for' loop below)
-			m_node = h_node.change_ext('.moc')
-			bld.node_deps[(self.inputs[0].parent.abspath(), m_node.name)] = h_node
-
-			# create the task
-			task = Task.classes['moc'](env=self.env, generator=self.generator)
-			task.set_inputs(h_node)
-			task.set_outputs(m_node)
-
-			# direct injection in the build phase (safe because called from the main thread)
-			gen = bld.producer
-			gen.outstanding.insert(0, task)
-			gen.total += 1
+			if not h_node:
+				raise Errors.WafError('No source found for %r which is a moc file' % d)
 
+			# create the moc task
+			task = self.create_moc_task(h_node, m_node)
 			moctasks.append(task)
 
-		# remove raw deps except the moc files to save space (optimization)
-		tmp_lst = bld.raw_deps[self.uid()] = mocfiles
-
-		# look at the file inputs, it is set right above
-		lst = bld.node_deps.get(self.uid(), ())
-		for d in lst:
-			name = d.name
-			if name.endswith('.moc'):
-				task = Task.classes['moc'](env=self.env, generator=self.generator)
-				task.set_inputs(bld.node_deps[(self.inputs[0].parent.abspath(), name)]) # 1st element in a tuple
-				task.set_outputs(d)
-
-				gen = bld.producer
-				gen.outstanding.insert(0, task)
-				gen.total += 1
-
-				moctasks.append(task)
-
 		# simple scheduler dependency: run the moc task before others
 		self.run_after.update(set(moctasks))
 		self.moc_done = 1
 
-	run = Task.classes['cxx'].__dict__['run']
-
 class trans_update(Task.Task):
 	"""Update a .ts files from a list of C++ files"""
 	run_str = '${QT_LUPDATE} ${SRC} -ts ${TGT}'
@@ -232,7 +261,7 @@ class XMLHandler(ContentHandler):
 def create_rcc_task(self, node):
 	"Create rcc and cxx tasks for *.qrc* files"
 	rcnode = node.change_ext('_rc.cpp')
-	rcctask = self.create_task('rcc', node, rcnode)
+	self.create_task('rcc', node, rcnode)
 	cpptask = self.create_task('cxx', rcnode, rcnode.change_ext('.o'))
 	try:
 		self.compiled_tasks.append(cpptask)
@@ -295,12 +324,12 @@ def apply_qt4(self):
 	for flag in self.to_list(self.env['CXXFLAGS']):
 		if len(flag) < 2: continue
 		f = flag[0:2]
-		if f in ['-D', '-I', '/D', '/I']:
+		if f in ('-D', '-I', '/D', '/I'):
 			if (f[0] == '/'):
 				lst.append('-' + flag[1:])
 			else:
 				lst.append(flag)
-	self.env['MOC_FLAGS'] = lst
+	self.env.append_value('MOC_FLAGS', lst)
 
 @extension(*EXT_QT4)
 def cxx_hook(self, node):
@@ -314,13 +343,14 @@ class rcc(Task.Task):
 	Process *.qrc* files
 	"""
 	color   = 'BLUE'
-	run_str = '${QT_RCC} -name ${SRC[0].name} ${SRC[0].abspath()} ${RCC_ST} -o ${TGT}'
+	run_str = '${QT_RCC} -name ${tsk.rcname()} ${SRC[0].abspath()} ${RCC_ST} -o ${TGT}'
 	ext_out = ['.h']
 
+	def rcname(self):
+		return os.path.splitext(self.inputs[0].name)[0]
+
 	def scan(self):
 		"""Parse the *.qrc* files"""
-		node = self.inputs[0]
-
 		if not has_xml:
 			Logs.error('no xml support was found, the rcc dependencies will be incomplete!')
 			return ([], [])
@@ -328,9 +358,11 @@ class rcc(Task.Task):
 		parser = make_parser()
 		curHandler = XMLHandler()
 		parser.setContentHandler(curHandler)
-		fi = open(self.inputs[0].abspath())
-		parser.parse(fi)
-		fi.close()
+		fi = open(self.inputs[0].abspath(), 'r')
+		try:
+			parser.parse(fi)
+		finally:
+			fi.close()
 
 		nodes = []
 		names = []
@@ -347,6 +379,10 @@ class moc(Task.Task):
 	"""
 	color   = 'BLUE'
 	run_str = '${QT_MOC} ${MOC_FLAGS} ${MOCCPPPATH_ST:INCPATHS} ${MOCDEFINES_ST:DEFINES} ${SRC} ${MOC_ST} ${TGT}'
+	def keyword(self):
+		return "Creating"
+	def __str__(self):
+		return self.outputs[0].path_from(self.generator.bld.launch_node())
 
 class ui4(Task.Task):
 	"""
@@ -385,6 +421,7 @@ def configure(self):
 	"""
 	self.find_qt4_binaries()
 	self.set_qt4_libs_to_check()
+	self.set_qt4_defines()
 	self.find_qt4_libraries()
 	self.add_qt4_rpath()
 	self.simplify_qt4_libs()
@@ -404,8 +441,8 @@ def find_qt4_binaries(self):
 
 	# the qt directory has been given from QT4_ROOT - deduce the qt binary path
 	if not qtdir:
-		qtdir = self.environ.get('QT4_ROOT', '')
-		qtbin = os.path.join(qtdir, 'bin')
+		qtdir = os.environ.get('QT4_ROOT', '')
+		qtbin = os.environ.get('QT4_BIN', None) or os.path.join(qtdir, 'bin')
 
 	if qtbin:
 		paths = [qtbin]
@@ -432,15 +469,15 @@ def find_qt4_binaries(self):
 	# keep the one with the highest version
 	cand = None
 	prev_ver = ['4', '0', '0']
-	for qmk in ['qmake-qt4', 'qmake4', 'qmake']:
+	for qmk in ('qmake-qt4', 'qmake4', 'qmake'):
 		try:
 			qmake = self.find_program(qmk, path_list=paths)
 		except self.errors.ConfigurationError:
 			pass
 		else:
 			try:
-				version = self.cmd_and_log([qmake, '-query', 'QT_VERSION']).strip()
-			except self.errors.ConfigurationError:
+				version = self.cmd_and_log(qmake + ['-query', 'QT_VERSION']).strip()
+			except self.errors.WafError:
 				pass
 			else:
 				if version:
@@ -453,9 +490,11 @@ def find_qt4_binaries(self):
 	else:
 		self.fatal('Could not find qmake for qt4')
 
-	qtbin = self.cmd_and_log([self.env.QMAKE, '-query', 'QT_INSTALL_BINS']).strip() + os.sep
+	qtbin = self.cmd_and_log(self.env.QMAKE + ['-query', 'QT_INSTALL_BINS']).strip() + os.sep
 
 	def find_bin(lst, var):
+		if var in env:
+			return
 		for f in lst:
 			try:
 				ret = self.find_program(f, path_list=paths)
@@ -467,20 +506,19 @@ def find_qt4_binaries(self):
 
 	find_bin(['uic-qt3', 'uic3'], 'QT_UIC3')
 	find_bin(['uic-qt4', 'uic'], 'QT_UIC')
-	if not env['QT_UIC']:
+	if not env.QT_UIC:
 		self.fatal('cannot find the uic compiler for qt4')
 
-	try:
-		uicver = self.cmd_and_log(env['QT_UIC'] + " -version 2>&1").strip()
-	except self.errors.ConfigurationError:
-		self.fatal('this uic compiler is for qt3, add uic for qt4 to your path')
+	self.start_msg('Checking for uic version')
+	uicver = self.cmd_and_log(env.QT_UIC + ["-version"], output=Context.BOTH)
+	uicver = ''.join(uicver).strip()
 	uicver = uicver.replace('Qt User Interface Compiler ','').replace('User Interface Compiler for Qt', '')
-	self.msg('Checking for uic version', '%s' % uicver)
+	self.end_msg(uicver)
 	if uicver.find(' 3.') != -1:
 		self.fatal('this uic compiler is for qt3, add uic for qt4 to your path')
 
 	find_bin(['moc-qt4', 'moc'], 'QT_MOC')
-	find_bin(['rcc'], 'QT_RCC')
+	find_bin(['rcc-qt4', 'rcc'], 'QT_RCC')
 	find_bin(['lrelease-qt4', 'lrelease'], 'QT_LRELEASE')
 	find_bin(['lupdate-qt4', 'lupdate'], 'QT_LUPDATE')
 
@@ -494,26 +532,28 @@ def find_qt4_binaries(self):
 
 @conf
 def find_qt4_libraries(self):
-	qtlibs = getattr(Options.options, 'qtlibs', '')
+	qtlibs = getattr(Options.options, 'qtlibs', None) or os.environ.get("QT4_LIBDIR", None)
 	if not qtlibs:
 		try:
-			qtlibs = self.cmd_and_log([self.env.QMAKE, '-query', 'QT_INSTALL_LIBS']).strip()
+			qtlibs = self.cmd_and_log(self.env.QMAKE + ['-query', 'QT_INSTALL_LIBS']).strip()
 		except Errors.WafError:
-			qtdir = self.cmd_and_log([self.env.QMAKE, '-query', 'QT_INSTALL_PREFIX']).strip() + os.sep
+			qtdir = self.cmd_and_log(self.env.QMAKE + ['-query', 'QT_INSTALL_PREFIX']).strip() + os.sep
 			qtlibs = os.path.join(qtdir, 'lib')
 	self.msg('Found the Qt4 libraries in', qtlibs)
 
-	qtincludes = self.cmd_and_log([self.env.QMAKE, '-query', 'QT_INSTALL_HEADERS']).strip()
+	qtincludes =  os.environ.get("QT4_INCLUDES", None) or self.cmd_and_log(self.env.QMAKE + ['-query', 'QT_INSTALL_HEADERS']).strip()
 	env = self.env
 	if not 'PKG_CONFIG_PATH' in os.environ:
 		os.environ['PKG_CONFIG_PATH'] = '%s:%s/pkgconfig:/usr/lib/qt4/lib/pkgconfig:/opt/qt4/lib/pkgconfig:/usr/lib/qt4/lib:/opt/qt4/lib' % (qtlibs, qtlibs)
 
 	try:
+		if os.environ.get("QT4_XCOMPILE", None):
+			raise self.errors.ConfigurationError()
 		self.check_cfg(atleast_pkgconfig_version='0.1')
 	except self.errors.ConfigurationError:
 		for i in self.qt4_vars:
 			uselib = i.upper()
-			if Utils.unversioned_sys_platform == "darwin":
+			if Utils.unversioned_sys_platform() == "darwin":
 				# Since at least qt 4.7.3 each library locates in separate directory
 				frameworkName = i + ".framework"
 				qtDynamicLib = os.path.join(qtlibs, frameworkName, i)
@@ -523,7 +563,7 @@ def find_qt4_libraries(self):
 				else:
 					self.msg('Checking for %s' % i, False, 'YELLOW')
 				env.append_unique('INCLUDES_' + uselib, os.path.join(qtlibs, frameworkName, 'Headers'))
-			elif sys.platform != "win32":
+			elif env.DEST_OS != "win32":
 				qtDynamicLib = os.path.join(qtlibs, "lib" + i + ".so")
 				qtStaticLib = os.path.join(qtlibs, "lib" + i + ".a")
 				if os.path.exists(qtDynamicLib):
@@ -599,7 +639,7 @@ def simplify_qt4_libs(self):
 def add_qt4_rpath(self):
 	# rpath if wanted
 	env = self.env
-	if Options.options.want_rpath:
+	if getattr(Options.options, 'want_rpath', False):
 		def process_rpath(vars_, coreval):
 			for d in vars_:
 				var = d.upper()
@@ -625,6 +665,15 @@ def set_qt4_libs_to_check(self):
 		self.qt4_vars_debug = [a + '_debug' for a in self.qt4_vars]
 	self.qt4_vars_debug = Utils.to_list(self.qt4_vars_debug)
 
+@conf
+def set_qt4_defines(self):
+	if sys.platform != 'win32':
+		return
+	for x in self.qt4_vars:
+		y = x[2:].upper()
+		self.env.append_unique('DEFINES_%s' % x.upper(), 'QT_%s_LIB' % y)
+		self.env.append_unique('DEFINES_%s_DEBUG' % x.upper(), 'QT_%s_LIB' % y)
+
 def options(opt):
 	"""
 	Command-line options
